/* ssh-keyserv.c
**
** Copyright (c) 1997 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
   
#include <stdio.h>
#include <syslog.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>


#ifndef SSH_KEYDIR
#define SSH_KEYDIR "/var/ssh-keys"
#endif

#ifndef SSH_KEYGEN
#define SSH_KEYGEN "/usr/local/bin/ssh-keygen"
#endif

static char *herr2str(int err)
{
    static char buf[1024];

    
    switch (err)
    {
      case HOST_NOT_FOUND:
	return "host not found";

      case TRY_AGAIN:
	return "try again";

      case NO_RECOVERY:
	return "no recovery";

      case NO_DATA:
	return "no data";

      default:
	sprintf(buf, "code #%d", err);
	return buf;
    }
}


static void make_keys(char *host)
{
    char buf[32768];
    FILE *fp;

    
    sprintf(buf,
	    "%s -b 1024 -N '' -f '%s/%s' -C 'root@%s'",
	    SSH_KEYGEN,
	    SSH_KEYDIR, host, host);
    
    fp = popen(buf, "r");
    if (fp == NULL)
    {
	syslog(LOG_ERR, "popen(%s): %m", buf);
	exit(1);
    }

    while (fgets(buf, sizeof(buf), fp))
	;

    pclose(fp);
}

static int valid_hostname(char *h)
{
    while (*h && (isalnum(*h) || *h == '-' || *h == '.'))
	++h;

    return (*h == 0);
}

int main(int argc, char *argv[])
{
    struct sockaddr_in sin;
    struct hostent *hp, *hp2;
    char buf[32768];
    int len, fd, nlen, i;


    openlog("ssh-keyserv", LOG_PID, LOG_AUTH);
    
    len = sizeof(sin);
    if (getpeername(0, (struct sockaddr *) &sin, &len) < 0)
    {
	syslog(LOG_ERR, "getpeername: %m");
	exit(1);
    }

    if (ntohs(sin.sin_port) >= 1024)
    {
	syslog(LOG_WARNING, "denied client %s:%d: high port",
	       inet_ntoa(sin.sin_addr),
	       ntohs(sin.sin_port));
	exit(1);
    }
    
    hp = gethostbyaddr((char *) &sin.sin_addr, sizeof(sin.sin_addr), AF_INET);
    if (hp == NULL || hp->h_name == NULL)
    {
	syslog(LOG_ERR, "gethostbyaddr(%s): %s",
	       inet_ntoa(sin.sin_addr),
	       herr2str(h_errno));
	exit(1);
    }
    
    if (strlen(hp->h_name) > 1024)
    {
	syslog(LOG_ERR, "hostname for IP# %s is too long",
	       inet_ntoa(sin.sin_addr));
	exit(1);
    }

    if (!valid_hostname(hp->h_name))
    {
	syslog(LOG_ERR, "hostname %s contains invalid characters",
	       hp->h_name);
	exit(1);
    }
    
    hp2 = gethostbyname(hp->h_name);
    if (hp2 == NULL || hp2->h_addr_list == NULL)
    {
	syslog(LOG_ERR, "gethostbyname(%s): %s",
	       hp->h_name,
	       herr2str(h_errno));
	exit(1);
    }

    for (i = 0; hp2->h_addr_list[i]; i++)
	if (memcmp(&sin.sin_addr,
		   hp2->h_addr_list[i],
		   sizeof(sin.sin_addr)) == 0)
	    break;

    if (hp2->h_addr_list[i] == NULL)
    {
	syslog(LOG_ERR, "%s == %s != %s - request denied",
	       inet_ntoa(sin.sin_addr),
	       hp->h_name,
	       inet_ntoa(sin.sin_addr));
	exit(1);
    }
    
    sprintf(buf, "%s/%s", SSH_KEYDIR, hp->h_name);
    fd = open(buf, O_RDONLY);
    if (fd == -1)
	make_keys(hp->h_name);
    
    fd = open(buf, O_RDONLY);
    if (fd == -1)
    {
	syslog(LOG_ERR, "%s: open(%s): %m", hp->h_name, buf);
	exit(1);
    }

    len = read(fd, buf, sizeof(buf));
    if (len < 0)
    {
	syslog(LOG_ERR, "%s: read(%s): %m", hp->h_name, buf);
	exit(1);
    }


    nlen = htonl(len);
    write(1, &nlen, sizeof(nlen));
    write(1, buf, len);
    close(fd);

    
    sprintf(buf, "%s/%s.pub", SSH_KEYDIR, hp->h_name);
    fd = open(buf, O_RDONLY);
    if (fd == -1)
    {
	syslog(LOG_ERR, "%s: open(%s): %m", hp->h_name, buf);
	exit(1);
    }

    len = read(fd, buf, sizeof(buf));
    if (len < 0)
    {
	syslog(LOG_ERR, "%s: read(%s): %m", hp->h_name, buf);
	exit(1);
    }


    nlen = htonl(len);
    write(1, &nlen, sizeof(nlen));
    write(1, buf, len);
    close(fd);

    exit(0);
}
